@@ -243,15 +243,13 @@ typedef struct getdns_upstream {
 	unsigned is_sync_loop : 1;
 
 	/* EDNS cookies */
-	uint32_t secret;
-	uint8_t  client_cookie[8];
-	uint8_t  prev_client_cookie[8];
-	uint8_t  server_cookie[32];
-
-	unsigned has_client_cookie : 1;
-	unsigned has_prev_client_cookie : 1;
-	unsigned has_server_cookie : 1;
-	unsigned server_cookie_len : 5;
+	uint8_t  server_cookie[40];
+	size_t   server_cookie_len;
+	
+	uint64_t                src_addr_checked;
+	struct sockaddr_storage src_addr;
+	socklen_t               src_addr_len;
+	char                    src_addr_str[INET6_ADDRSTRLEN];
 
 	/* TSIG */
 	uint8_t          tsig_dname[256];
@@ -327,6 +325,7 @@ struct getdns_context {
 	size_t               namespace_count;
 	uint64_t             timeout;
 	uint64_t             idle_timeout;
+	int                  tcp_send_timeout; /* -1 is unset */
 	getdns_redirects_t   follow_redirects;
 	getdns_list          *dns_root_servers;
 
